/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */


/*
 * conn8.c --
 *
 *      Tcl 8 version of some Conn routines
 */

static const char *RCSID = "@(#) $Header: /usr/local/cvsroot/nsd/aolserver/nsd/conn8.c,v 1.8 2001/06/11 06:08:09 mayoff Exp $, compiled: " __DATE__ " " __TIME__;

#define USE_TCL8X

#include "nsd.h"

#include <ctype.h>

#define WHITESPACE             " \t\r\n"
#define DEFAULT_CHARSET_CONFIG "OutputCharset"
#define HACK_CONTENT_TYPE_CONFIG "HackContentType"
#define CHARSET_CONFIGSECTION  "ns/charsets"
#define DEFAULT_CHARSET        "iso-8859-1"

/*
 * This is truly ugly, but we need these functions to be able to set
 * the Tcl library path so we can find our encodings.
 */
extern void TclSetLibraryPath(Tcl_Obj *pathPtr);
extern Tcl_Obj *TclGetLibraryPath(void);

/*
 * Local functions defined in this file.
 */

static void AddCharsetEncoding(char *charsetName, char *encodingName);
static Tcl_Encoding GetEncodingForType(char *contentType, int *usedDefaultCharsetP);
static Ns_TclInterpInitProc CharsetInterpInit;
static Ns_TclInterpInitProc CharsetInterpInit2;
static Tcl_CmdProc EncodingForCharsetCmd;
static void FindEncodings(void);

/*
 * Local variables defined in this file.
 */

static char *defaultCharset;
static Tcl_Encoding defaultEncoding;
static int hackContentTypeP;
static Tcl_Obj *savedTclLibrary;

/*
 * charset2EncodingMap maps IANA (and some non-standard) charset
 * names to Tcl_Encodings.
 */
static Tcl_HashTable charset2EncodingMap;

/*
 * Tcl-format list of charsets for which we have encodings, for the
 * ns_charsets command.
 */
static char *charsetList;

/*
 * The default encoding name map, to initialize charset2EncodingMap.
 */
static struct charset2EncodingPair {
    char *charset;
    char *encoding;
} charset2EncodingTable[] = {

    { "identity", "identity" },
    { "8bit", "8bit" },
    { "ascii", "ascii" },
    { "big5", "big5" },
    { "euc-cn", "euc-cn" },
    { "euc-jp", "euc-jp" },
    { "euc-kr", "euc-kr" },
    { "gb12345", "gb12345" },
    { "gb1988", "gb1988" },
    { "gb2312", "gb2312" },
    { "ibm437", "cp437" },
    { "ibm737", "cp737" },
    { "ibm775", "cp775" },
    { "ibm850", "cp850" },
    { "ibm852", "cp852" },
    { "ibm855", "cp855" },
    { "ibm857", "cp857" },
    { "ibm860", "cp860" },
    { "ibm861", "cp861" },
    { "ibm862", "cp862" },
    { "ibm863", "cp863" },
    { "ibm864", "cp864" },
    { "ibm865", "cp865" },
    { "ibm866", "cp866" },
    { "ibm869", "cp869" },
    { "ibm874", "cp874" },
    { "ibm932", "cp932" },
    { "ibm936", "cp936" },
    { "ibm949", "cp949" },
    { "ibm950", "cp950" },
    { "iso-2022-jp", "iso2022-jp" },
    { "iso-2022-kr", "iso2022-kr" },
    { "iso-8859-1", "iso8859-1" },
    { "iso-8859-2", "iso8859-2" },
    { "iso-8859-3", "iso8859-3" },
    { "iso-8859-4", "iso8859-4" },
    { "iso-8859-5", "iso8859-5" },
    { "iso-8859-6", "iso8859-6" },
    { "iso-8859-7", "iso8859-7" },
    { "iso-8859-8", "iso8859-8" },
    { "iso-8859-9", "iso8859-9" },
    { "jis0201", "jis0201" },
    { "jis0208", "jis0208" },
    { "jis0212", "jis0212" },
    { "koi8-r", "koi8-r" },
    { "korean", "ksc5601" },
    { "ksc_5601", "ksc5601" },
    { "mac", "macRoman" },
    { "mac-centeuro", "macCentEuro" },
    { "mac-centraleupore", "macCentEuro" },
    { "mac-croatian", "macCroatian" },
    { "mac-cyrillic", "macCyrillic" },
    { "mac-greek", "macGreek" },
    { "mac-iceland", "macIceland" },
    { "mac-japan", "macJapan" },
    { "mac-roman", "macRoman" },
    { "mac-romania", "macRomania" },
    { "mac-thai", "macThai" },
    { "mac-turkish", "macTurkish" },
    { "mac-ukraine", "macUkraine" },
    { "maccenteuro", "macCentEuro" },
    { "maccentraleupore", "macCentEuro" },
    { "maccroatian", "macCroatian" },
    { "maccyrillic", "macCyrillic" },
    { "macgreek", "macGreek" },
    { "maciceland", "macIceland" },
    { "macintosh", "macRoman" },
    { "macjapan", "macJapan" },
    { "macroman", "macRoman" },
    { "macromania", "macRomania" },
    { "macthai", "macThai" },
    { "macturkish", "macTurkish" },
    { "macukraine", "macUkraine" },
    { "shift_jis", "shiftjis" },
    { "unicode", "unicode" },
    { "us-ascii", "ascii" },
    { "utf-8", "utf-8" },
    { "windows-1250", "cp1250" },
    { "windows-1251", "cp1251" },
    { "windows-1252", "cp1252" },
    { "windows-1253", "cp1253" },
    { "windows-1254", "cp1254" },
    { "windows-1255", "cp1255" },
    { "windows-1256", "cp1256" },
    { "windows-1257", "cp1257" },
    { "windows-1258", "cp1258" },
    { "x-mac", "macRoman" },
    { "x-mac-centeuro", "macCentEuro" },
    { "x-mac-centraleupore", "macCentEuro" },
    { "x-mac-croatian", "macCroatian" },
    { "x-mac-cyrillic", "macCyrillic" },
    { "x-mac-greek", "macGreek" },
    { "x-mac-iceland", "macIceland" },
    { "x-mac-japan", "macJapan" },
    { "x-mac-roman", "macRoman" },
    { "x-mac-romania", "macRomania" },
    { "x-mac-thai", "macThai" },
    { "x-mac-turkish", "macTurkish" },
    { "x-mac-ukraine", "macUkraine" },
    { "x-macintosh", "macRoman" },

    { NULL, NULL }
};

/*
 *==========================================================================
 * API functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * Ns_WriteConn --
 *
 *	This will write a buffer to the conn. It promises to write 
 *	all of it.
 *
 *      If we think we are writing the headers (which is the default),
 *      then we send the data exactly as it is given to us.  If we are
 *      truly in the headers, then they are supposed to be US-ASCII,
 *      which is a subset of UTF-8, so no translation should be needed
 *      if the user has been good and not put any 8-bit characters
 *      into it.
 *
 *      If we have been told that we are sending the content, and we
 *      have been given an encoding to translate the content to, then
 *      we assume that the caller is handing us UTF-8 bytes and we
 *      translate them to the preset encoding.
 *
 * Results:
 *	NS_OK/NS_ERROR
 *
 * Side effects:
 *	Stuff may be written 
 *
 *----------------------------------------------------------------------
 */

int
Ns_WriteConn(Ns_Conn *conn, char *buf, int len)
{
    int             nwrote;
    int             status;
    char           *utfBytes;
    int             utfCount; /* # of bytes in utfBytes */
    int             utfConvertedCount;  /* # of bytes of utfBytes converted */
    char            encodedBytes[BUFSIZ];
    int             encodedCount; /* # of bytes converted in encodedBytes */
    Tcl_Interp     *interp;
    Conn           *connPtr = (Conn *)conn;

    status = NS_OK;

    if (connPtr->sendState == Headers || connPtr->outputEncoding == NULL) {

	/*
         * Headers must be in US-ASCII.  Since US-ASCII is a subset of
	 * UTF-8, we won't bother encodnig anything.  It's up to
	 * the AOLserver user to only put US-ASCII into the headers.
	 *
	 * If we're in the content section, but we never figured out
	 * an encoding to use, then we just send them the raw data.
	 * This happens for binary data like images.
	 */

	status = NsWriteConnRaw(conn, buf, len);

    } else {

	/* connPtr->sendState == Content && connPtr->outputEncoding != NULL */

	utfBytes = buf;
	utfCount = len;
	interp = Ns_GetConnInterp(conn);

	while (utfCount > 0 && status == NS_OK) {

	    /* Convert a chunk to the desired encoding. */

	    status = Tcl_UtfToExternal(interp,
		connPtr->outputEncoding,
		utfBytes, utfCount,
		0, NULL,              /* flags, encoding state */
		encodedBytes, sizeof encodedBytes,
		&utfConvertedCount,
		&encodedCount,
		NULL                  /* # of chars encoded */
	    );

	    if (status != TCL_OK && status != TCL_CONVERT_NOSPACE) {
		status = NS_ERROR;
		break;
	    }

	    /* Send the converted chunk. */

	    status = NS_OK;
	    buf = encodedBytes;
	    len = encodedCount;

	    status = NsWriteConnRaw(conn, buf, len);

	    utfCount -= utfConvertedCount;
	    utfBytes += utfConvertedCount;
	}
    }

    return status;
}


/* 
 *==========================================================================
 * Exported functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * NsTclCharsetsCmd --
 *
 *      Tcl command to get the list of charsets for which we have encodings.
 *
 * Results:
 *	TCL_OK
 *
 * Side effects:
 *	Sets Tcl interpreter result.
 *
 *----------------------------------------------------------------------
 */

int
NsTclCharsetsCmd(ClientData dummy, Tcl_Interp *interp, int argc,
    char *argv[])
{
    Tcl_HashEntry  *entry;
    Tcl_HashSearch  search;
    Tcl_DString     ds;

    Tcl_SetResult(interp, charsetList, TCL_STATIC);

    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclStartContentCmd --
 *
 *      Set connPtr->sendState to "Content" and set the charset/encoding
 *      to use for further data.
 *
 * Results:
 *	NS_OK/NS_ERROR
 *
 * Side effects:
 *	connPtr->sendState and connPtr->encoding may be set.
 *
 *----------------------------------------------------------------------
 */

int
NsTclStartContentCmd(ClientData dummy, Tcl_Interp *interp, int argc,
    char *argv[])
{
    Tcl_Encoding  encoding;
    Conn         *connPtr;
    int           status;
    int           i;

    status = TCL_OK;
    encoding = NULL;

    for (i = 1; i < argc && status == TCL_OK; i++) {
	if (strcmp(argv[i], "-charset") == 0) {

	    if (encoding != NULL) {
		Tcl_AppendResult(interp, argv[0],
		    ": charset may only be specified by one flag", NULL);
		status = TCL_ERROR;
	    }

	    if (++i >= argc) {
		Tcl_AppendResult(interp, argv[0],
		    ": missing argument for -charset flag", NULL);
		status = TCL_ERROR;
	    }

	    encoding = (Tcl_Encoding) NsGetCharsetEncoding(argv[i]);
	    if (encoding == NULL) {
		Tcl_AppendResult(interp, argv[0],
		    ": could not find an encoding for charset ",
		    argv[i], NULL);
		status = TCL_ERROR;
	    }
	}

	else if (strcmp(argv[i], "-type") == 0) {

	    if (encoding != NULL) {
		Tcl_AppendResult(interp, argv[0],
		    ": charset may only be specified by one flag", NULL);
		status = TCL_ERROR;
	    }

	    if (++i >= argc) {
		Tcl_AppendResult(interp, argv[0],
		    ": missing argument for -type flag", NULL);
		status = TCL_ERROR;
	    }

	    encoding = GetEncodingForType(argv[i], NULL);
	}

	else {
	    Tcl_AppendResult(interp, "usage: ", argv[0],
		" ?-charset charsetname? ?-type content-type?", NULL);
	    status = TCL_ERROR;
	}
    }

    if (status != TCL_OK) {
	return status;
    }

    connPtr = (Conn *)Ns_TclGetConn(interp);
    connPtr->sendState = Content;
    connPtr->outputEncoding = encoding;

    return status;
}


/*
 *----------------------------------------------------------------------
 *
 * NsCharsetInit --
 *
 *      Initialize charset translation support.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	charset2EncodingMap is initialized from built-in table and
 *      nsd config.  charsetList is initialized as well.
 *
 *----------------------------------------------------------------------
 */

void
NsCharsetInit(void)
{
    int             i;
    Tcl_HashEntry  *entry;
    int             new_p;
    Ns_Set         *section;
    Tcl_HashSearch  search;
    Tcl_DString     ds;
    char           *charset;

    Ns_TclInitInterps("", CharsetInterpInit, NULL);

    /* Initialize charset2EncodingMap. */

    Tcl_InitHashTable(&charset2EncodingMap, TCL_STRING_KEYS);

    for (i = 0; charset2EncodingTable[i].charset != NULL; i++) {
	AddCharsetEncoding(charset2EncodingTable[i].charset,
	    charset2EncodingTable[i].encoding);
    }

    section = Ns_ConfigGetSection(CHARSET_CONFIGSECTION);
    if (section != NULL) {
	for (i = 0; i < Ns_SetSize(section); i++) {
	    AddCharsetEncoding(Ns_SetKey(section, i),
		Ns_SetValue(section, i));
	}
    }

    /* Initialize charsetList. */

    Tcl_DStringInit(&ds);

    entry = Tcl_FirstHashEntry(&charset2EncodingMap, &search);
    while (entry != NULL) {
	Tcl_DStringAppendElement(&ds,
	    (char *) Tcl_GetHashKey(&charset2EncodingMap, entry));
	entry = Tcl_NextHashEntry(&search);
    }

    charsetList = Ns_StrDup(Tcl_DStringValue(&ds));

    Tcl_DStringFree(&ds);

    /* Initialize default output charset. */

    defaultCharset = Ns_ConfigGetValue(NS_CONFIG_PARAMETERS,
	DEFAULT_CHARSET_CONFIG);
    if (defaultCharset == NULL) {
	defaultCharset = DEFAULT_CHARSET;
    }

    defaultEncoding = NsGetCharsetEncoding(defaultCharset);
    if (defaultEncoding == NULL) {
	Ns_Fatal("could not find encoding for default output charset \"%s\"",
	    defaultCharset);
    }

    /* Initialize hackContentTypeP. */

    hackContentTypeP = NS_TRUE;
    Ns_ConfigGetBool(NS_CONFIG_PARAMETERS, HACK_CONTENT_TYPE_CONFIG,
	&hackContentTypeP);

    /* Finish up. */

    Ns_TclInitInterps("", CharsetInterpInit2, NULL);
}


/*
 *----------------------------------------------------------------------
 *
 * NsConnComputeEncoding --
 *
 *      Figure out what charset the client will expect, based on the
 *      output headers, and set the appropriate outputEncoding.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	connPtr->outputEncoding may be set.
 *
 *----------------------------------------------------------------------
 */

void
NsConnComputeEncoding(Ns_Conn *conn)
{
    Conn *connPtr = (Conn *) conn;
    int index;
    char *contentType;

    index = Ns_SetIFind(conn->outputheaders, "Content-Type");
    if (index < 0)
	return;

    contentType = Ns_SetValue(conn->outputheaders, index);

    if (contentType != NULL) {
	int usedDefaultCharsetP;
	connPtr->outputEncoding = GetEncodingForType(contentType,
	    &usedDefaultCharsetP);

	if (usedDefaultCharsetP && hackContentTypeP) {
	    Ns_DString ds;

	    Ns_DStringInit(&ds);
	    Ns_DStringAppend(&ds, contentType);
	    Ns_DStringAppend(&ds, "; charset=");
	    Ns_DStringAppend(&ds, defaultCharset);
	    Ns_SetPutValue(conn->outputheaders, index, Ns_DStringValue(&ds));
	}
    }
}


/*
 *----------------------------------------------------------------------
 *
 * GetEncodingForType --
 *
 *      Figure out what charset the client will expect, based on the
 *      content-type.
 *
 *      The rules are:
 *
 *      1. If there is no Content-Type header, then we will not use
 *         any encoding and the content will be sent without translation.
 *
 *      2. If the Content-Type header contains a charset parameter,
 *         then we look up the charset.  If we find an encoding, 
 *         we will use it.  If we don't find an encoding, we log a
 *         message and return NS_ERROR.
 *
 *      3. If the Content-Type header contains no charset parameter,
 *         but specifies a text type, then we assume that the charset
 *         is iso-8859-1 (as required by RFC 2616 section 3.7.1).
 *
 *      4. Otherwise, we will not use any encoding and the content will
 *         be sent without translation.
 *
 * Results:
 *	A Tcl_Encoding, or NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Tcl_Encoding
GetEncodingForType(char *contentType, int *usedDefaultCharsetP)
{
    Tcl_Encoding  encoding = NULL;
    char         *charset = NULL;
    char         *p;

    /* Skip leading whitespace. */
    contentType += strspn(contentType, WHITESPACE);

    p = contentType;
    charset = NULL;

    do {

	p = strchr(p, ';');                  /* Find a semicolon. */
	if (p == NULL) break;

	p++;                                 /* Pass the semicolon. */
	p += strspn(p, WHITESPACE);          /* Skip whitespace. */

	if (strncasecmp(p, "charset", 7) != 0) continue;
	p += 7;                              /* Pass "charset". */

	p += strspn(p, WHITESPACE);          /* Skip whitespace. */
	if (*p != '=') continue;             /* Find an "="? */

	p++;                                 /* Pass the "=". */
	charset = p + strspn(p, WHITESPACE); /* Skip whitespace. */

    } while (charset == NULL);

    if (charset == NULL && strncasecmp(contentType, "text/", 5) == 0) {
	charset = defaultCharset;
	if (usedDefaultCharsetP != NULL) {
	    *usedDefaultCharsetP = NS_TRUE;
	}
    }

    else if (usedDefaultCharsetP != NULL) {
	*usedDefaultCharsetP = NS_FALSE;
    }

    if (charset != NULL) {
	Ns_DString ds;
	Ns_DStringInit(&ds);
	Ns_DStringAppend(&ds, charset);

	for (p = Ns_DStringValue(&ds); *p != '\0'; p++) {
	    *p = tolower(*p);
	}

	encoding = (Tcl_Encoding) NsGetCharsetEncoding(Ns_DStringValue(&ds));

	if (encoding == NULL) {
	    Ns_Log(Warning, "no encoding found for charset \"%s\"",
		charset);
	}

	Ns_DStringFree(&ds);
    }

    return encoding;
}


/*
 *----------------------------------------------------------------------
 *
 * NsConnEncodeForOutput --
 *
 *      Translate the buffer to the output encoding for the connection.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets ds.
 *
 *----------------------------------------------------------------------
 */

void
NsConnEncodeForOutput(Ns_Conn *conn, char *buf, int len, Tcl_DString *ds)
{
    Conn *connPtr = (Conn *) conn;

    if (len < 0) {
	len = strlen(buf);
    }

    Tcl_UtfToExternalDString(connPtr->outputEncoding, buf, len, ds);
}


/*
 *----------------------------------------------------------------------
 *
 * NsGetCharsetEncoding --
 *
 *      Get the Tcl_Encoding corresponding to a charset.
 *
 * Results:
 *	Tcl_Encoding or NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void *
NsGetCharsetEncoding(char *charset)
{
    Tcl_HashEntry *entry;
    Tcl_Encoding   encoding;
    char          *p;
    Ns_DString     ds;

    Ns_DStringInit(&ds);
    Ns_DStringAppend(&ds, charset);

    for (p = Ns_DStringValue(&ds); *p != '\0'; p++) {
	*p = tolower(*p);
    }

    entry = Tcl_FindHashEntry(&charset2EncodingMap, Ns_DStringValue(&ds));

    Ns_DStringFree(&ds);

    return (entry == NULL) ? NULL : (Tcl_Encoding) Tcl_GetHashValue(entry);
}


/*
 *==========================================================================
 * Static functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * AddCharsetEncoding --
 *
 *      Add a charset/encoding pair to charset2EncodingMap, if the
 *      encoding name can be resolved to a Tcl_Encoding.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	An entry may be added to charset2EncodingMap.
 *
 *----------------------------------------------------------------------
 */

static void
AddCharsetEncoding(char *charsetName, char *encodingName)
{
    Tcl_Encoding   encoding;
    Tcl_HashEntry *entry;
    int            new_p;

    encoding = Tcl_GetEncoding(NULL, encodingName);
    if (encoding == NULL) {
	Ns_Log(Warning,
	    "could not load Tcl encoding \"%s\" for charset \"%s\"",
	    encodingName, charsetName);
	return;
    }

    entry = Tcl_CreateHashEntry(&charset2EncodingMap, charsetName, &new_p);
    if (!new_p) {
	Tcl_FreeEncoding((Tcl_Encoding) Tcl_GetHashValue(entry));
    }
    Tcl_SetHashValue(entry, (ClientData) encoding);
}


/*
 *----------------------------------------------------------------------
 *
 * EncodingForCharsetCmd --
 *
 *      Return the name of the encoding for the specified charset.
 *
 * Results:
 *	Tcl result contains an encoding name or "".
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
EncodingForCharsetCmd(ClientData dummy, Tcl_Interp *interp, int argc,
    char **argv)
{
    Tcl_Encoding encoding;

    if (argc != 2) {
	Tcl_AppendResult(interp, "usage: ", argv[0], " charset", NULL);
	return TCL_ERROR;
    }

    encoding = (Tcl_Encoding) NsGetCharsetEncoding(argv[1]);
    if (encoding == NULL) {
	return TCL_OK;
    }

    Tcl_SetResult(interp, Tcl_GetEncodingName(encoding), TCL_VOLATILE);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * CharsetInterpInit --
 *
 *      Ensures that Tcl can find its encoding library, and adds
 *      charset-related commands to the parent interpreter.
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
CharsetInterpInit(Tcl_Interp *interp, void *content)
{
    Tcl_Obj  *newLibrary;
    Tcl_Obj  *newdir;
    int       objc;
    Tcl_Obj **objv;

    /*
     * Put nsd home at the front of the library path so it will
     * find the encoding files.
     */

    savedTclLibrary = TclGetLibraryPath();
    Tcl_IncrRefCount(savedTclLibrary);

    newdir = Tcl_NewStringObj(Ns_InfoHome(), -1);
    Tcl_AppendToObj(newdir, "/lib/tcl8.3", -1);
    newLibrary = Tcl_NewListObj(1, &newdir);

    Tcl_ListObjGetElements(NULL, savedTclLibrary, &objc, &objv);
    Tcl_ListObjReplace(NULL, newLibrary, 1, 0, objc, objv);

    TclSetLibraryPath(newLibrary);

    /*
     * Create the Tcl commands.
     */

    Tcl_CreateCommand(interp, "ns_charsets", NsTclCharsetsCmd,
	(ClientData) NULL, NULL);
    Tcl_CreateCommand(interp, "ns_startcontent", NsTclStartContentCmd,
	(ClientData) NULL, NULL);
    Tcl_CreateCommand(interp, "ns_encodingforcharset",
	EncodingForCharsetCmd, (ClientData) NULL, NULL);

    return NS_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * CharsetInterpInit2 --
 *
 *      Restore the Tcl library.
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
CharsetInterpInit2(Tcl_Interp *interp, void *content)
{
    TclSetLibraryPath(savedTclLibrary);
    Tcl_DecrRefCount(savedTclLibrary);
    savedTclLibrary = NULL;

    return NS_OK;
}

