/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.com/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */


/* 
 * urlencode.c --
 *
 *	Encode and decode URLs, as described in RFC 1738.
 */

static const char *RCSID = "@(#) $Header: /usr/local/cvsroot/nsd/aolserver/nsd/urlencode8.c,v 1.3 2000/07/14 16:55:39 mayoff Exp $, compiled: " __DATE__ " " __TIME__;

#define USE_TCL8X

#include "nsd.h"

#define DEFAULT_CHARSET "iso-8859-1"

/*
 * Local functions defined in this file
 */

static int HexDigit(char c);
static Tcl_Encoding GetUrlEncoding(char *charset);
static Ns_TclInterpInitProc EncodingInterpInit;

/*
 * Static variables defined in this file.
 */

static Tcl_Encoding defaultEncoding;

/*
 *==========================================================================
 * API functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 *
 * Ns_EncodeUrlCharset --
 *
 *	Take a URL and encode any non-alphanumeric characters into 
 *	%hexcode, according to the specified charset.
 *
 * Results:
 *	A pointer to the encoded string (which is part of the 
 *	passed-in DString's memory) 
 *
 * Side effects:
 *	Will append to the passed-in dstring 
 *
 *----------------------------------------------------------------------
 */

char *
Ns_EncodeUrlCharset(Ns_DString *pds, char *string, char *charset)
{
    Tcl_DString  ds;
    int          i;
    Tcl_Encoding encoding = GetUrlEncoding(charset);

    static char safechars[] =
	"abcdefghijklmnopqrstuvwxyz"
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	"0123456789";

    string = Tcl_UtfToExternalDString(encoding, string, -1, &ds);

    for (i = Tcl_DStringLength(&ds); i > 0; i--) {

        if (strchr(safechars, *string) != NULL) {
            Ns_DStringNAppend(pds, string, 1);
        } else {
            char buf[4];
            sprintf(buf, "%%%02x", (unsigned char) *string);
	    Ns_DStringNAppend(pds, buf, 3);
        }

        ++string;
    }

    Tcl_DStringFree(&ds);

    return pds->string;
}


/*
 *----------------------------------------------------------------------
 *
 * Ns_DecodeUrlCharset --
 *
 *	Take an encoded URL (with %hexcode, etc.) and decode it into 
 *	the DString, according to the specified charset.
 *
 * Results:
 *	A pointer to the dstring's memory, containing the decoded 
 *	URL, or NULL if incorrectly encoded.
 *
 * Side effects:
 *	Will append stuff to pds.
 *
 *----------------------------------------------------------------------
 */

char *
Ns_DecodeUrlCharset(Ns_DString *pds, char *string, char *charset)
{
    char         *copy;
    int           length;
    char         *p;
    char         *q;
    char          twobytes[3];
    Tcl_DString   ds;
    Tcl_Encoding  encoding = GetUrlEncoding(charset);

    length = strlen(string);
    copy = Ns_Malloc(length+1);
    twobytes[2] = '\0';

    p = string;
    q = copy;

    while (*p != '\0') {
	if (*p == '%' && HexDigit(p[1]) && HexDigit(p[2])) {
	    twobytes[0] = p[1];
	    twobytes[1] = p[2];
	    *q++ = (char) strtol(twobytes, (char **) NULL, 16);
	    p += 3;
	    length -= 2;
	} else {
	    *q++ = *p++;
	}
    }

    *q = '\0';

    Tcl_ExternalToUtfDString(encoding, copy, length, &ds);

    Ns_DStringAppend(pds, Tcl_DStringValue(&ds));

    Tcl_DStringFree(&ds);
    Ns_Free(copy);

    return pds->string;
}


/* 
 *==========================================================================
 * Exported functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * NsInitEncoding --
 *
 *	Initialize character encodings for Ns_DecodeUrl and
 *      Ns_EncodeUrl.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
NsEncodingInit(void)
{
    char *charset;

    charset = Ns_ConfigGetValue(NS_CONFIG_PARAMETERS, "URLCharset");
    if (charset == NULL) {
	charset = DEFAULT_CHARSET;
    }

    defaultEncoding = NsGetCharsetEncoding(charset);
    if (defaultEncoding == NULL) {
	Ns_Fatal("could not load encoding for default URL charset \"%s\"",
	    charset);
    }

    Ns_TclInitInterps("", EncodingInterpInit, NULL);
}


/*
 *----------------------------------------------------------------------
 *
 * NsTclURLCharsetCmd --
 *
 *	Implements ns_urlcharset. 
 *
 * Results:
 *	Tcl result. 
 *
 * Side effects:
 *	Sets connPtr->urlEncoding.
 *
 *----------------------------------------------------------------------
 */

int
NsTclURLCharsetCmd(ClientData dummy, Tcl_Interp *interp, int argc,
    char **argv)
{
    Tcl_Encoding  encoding;
    char         *charset;

    if (argc != 2) {
        Tcl_AppendResult(interp, "bad usage: should be \"",
                         argv[0], " charset", NULL);
        return TCL_ERROR;
    }

    charset = argv[1];
    encoding = NsGetCharsetEncoding(charset);

    if (encoding == NULL) {
	Tcl_AppendResult(interp, "no encoding for charset \"",
	    charset, "\"", NULL);
	return TCL_ERROR;
    }

    ((Conn *) Ns_GetConn())->urlEncoding = encoding;

    return TCL_OK;
}


/*
 *==========================================================================
 * Static functions
 *==========================================================================
 */

/*
 *----------------------------------------------------------------------
 *
 * HexDigit --
 *
 *	Determines if a character is a valid hexadecimal digit.
 *
 * Results:
 *	Boolean: true if char is a valid hex digit.
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

static int
HexDigit(char c)
{
    static char     hexdigits[] = "0123456789ABCDEFabcdef";

    return (c != '\0' && (strchr(hexdigits, c) != NULL));
}


/*
 *----------------------------------------------------------------------
 *
 * GetUrlEncoding --
 *
 *	Get the encoding to use in Ns_EncodeUrl/Ns_DecodeUrl.
 *
 * Results:
 *	A Tcl_Encoding.
 *
 * Side effects:
 *	None. 
 *
 *----------------------------------------------------------------------
 */

static Tcl_Encoding
GetUrlEncoding(char *charset)
{
    Tcl_Encoding  encoding = NULL;

    if (charset != NULL) {
	encoding = (Tcl_Encoding) NsGetCharsetEncoding(charset);
	if (encoding == NULL) {
	    Ns_Log(Warning, "no encoding found for charset \"%s\"",
		charset);
	}
    }

    if (encoding == NULL) {
	Conn *connPtr = (Conn *) Ns_GetConn();
	if (connPtr != NULL) {
	    encoding = connPtr->urlEncoding;
	}
    }

    if (encoding == NULL) {
	encoding = defaultEncoding;
    }

    return encoding;
}


/*
 *----------------------------------------------------------------------
 *
 * EncodingInterpInit --
 *
 *      Add url-encoding-related commands to the parent interpreter.
 *
 * Results:
 *	NS_OK.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
EncodingInterpInit(Tcl_Interp *interp, void *content)
{
    Tcl_CreateCommand(interp, "ns_urlcharset", NsTclURLCharsetCmd,
	(ClientData) NULL, NULL);

    return NS_OK;
}

